-- main controller loop 
-- v3.5 6/2/2021
-- Copyright (C) 2018-2021 dev

local module = {}

MMStatus = ""   --  global

function getWifiStatus()
  local vIP = wifi.sta.getip()
  if vIP == nil then
    vIP = "nil"
  end
  
  local _,_,d4 = vIP:find("%d*.%d*.%d*.(%d*)")
  --print(d)
  if d4 == nil then
    d4 = "nil"  
  end
  
  local vRssi = wifi.sta.getrssi()
  if vRssi == nil then
    vRssi = "nil"
  end

  return d4..":"..vRssi
--  return vRssi
end

local function slowLog(vRatio)
  local n = 1
  local v1 = vRatio
  while v1 >= 10 do
    v1 = v1 * .1
    n = n + 1
  end
  while v1 < 1.0 do
    v1 = v1 * 10
    n = n - 1
  end
  local y = (v1 - 1) / (v1 + 1)
  local y2 = y * y

  local sum = 0
  for i = 1, 15, 2 do
    sum = sum + 2 * (y / i)
    y = y * y2
  end

  local Exp = (n - 1) * 2.30258509299
  local Val = Exp + sum

--  print(Val, sum, Exp)
  return Val
end


local function calcCent(vRatio)
  local vm = { 3.473, 0.3266, 0.0521, 0.01246, 0.00394, 0.001543 }
  local a = { 0.003386128, 0.003354017, 0.003352514, 0.003351268, 0.003326146, 0.003324011 }
  local b = { 0.000214219, 0.000238792, 0.000232086, 0.000230287, 0.000220713, 0.000219557 }
  local j = 0
  for i = 1, 6 do
    if vRatio > vm[i] then
      j = i
      break
    end
  end
  if j == 0 then
    return 999.9
  end
--  logR = math.log(vRatio)
  local logR = slowLog(vRatio)
  local sumK = a[j] + b[j] * logR
  local degK = 1 / sumK
  local degC = degK - 273.15
  -- print(vRatio,a,b,logR, sumK, degK, degC)
  return degC
end


-- Main controller loop

local vState = 0 -- sOFF
local vStateo = 7 --sFAULT
local vStErr = 0 -- sOFF
local vAdc = 0
local vAdcOpen = 0
local vAdcOld = -1

local vR1 = 18823.52941 -- 24592.66097   -- parallel 27k (26640) and 220k/100k
local vR25 = 200000

local tick = 0
local tkFan = 0
local tkLED = 50
local tkState = 0
local sLed = 0

local tkIgn = 0		--	dev2020-07-02

gpio.mode(0, gpio.OUTPUT)	--pLED
gpio.mode(1, gpio.OUTPUT)	--pIGN was 3
--gpio.mode(4, gpio.OUTPUT)	--pBLED
gpio.mode(5, gpio.OPENDRAIN)	--pTHERMLO
gpio.mode(6, gpio.OUTPUT)	--pFAN
gpio.mode(8, gpio.OUTPUT)	--pGAS
gpio.mode(7, gpio.INPUT)	--pSWCH

gpio.write(5, 0)	-- pTHERMLO


local vSW = 0
local vSWo = 0

local vFan = 0
local vIgn = 0
local vGas = 0

local degC1 = 0
local degCOld = 0
local degOk = 70	--v3.2 70 was 50
local deg2Hot = 140


local vMsgOld = ""
local vStartT = 25	-- starting temperature for states 3,4,5

local vIgnOn = 0
local vGasOff = 0

local function  MM_doCommandLocal(topic, data)
  local vS = topic..": "..data
  print(vS)
  if vMsgOld == data then
    send_message("ok: "..vS)
    return
  else
    send_message("rxd: "..vS)
  end

  vMsgOld = data

  local t={} ; i=1
  for str in string.gmatch(data, "([^%s]+)") do
    t[i] = str
    i = i + 1
  end

  if t[1] then
    data = t[1]
  end

  local arg = t[2]
  local narg = arg and tonumber(arg) or 0

  print(data, arg, narg)

  if data == "MMclear" then
    vState = 1 --sCLR
    vStErr = 0
  elseif data == "MMlow" then
    vState = 2 --sLOW
    vStErr = 0
  elseif data == "MMrise" then
    vState = 3 --sRISE
    vStErr = 0
  elseif data == "MMcool" then
    vState = 6 --sCOOL
    vStErr = narg	-- v3.2 was 0
  elseif data == "MMi" and arg then
    vIgnOn = narg
  elseif data == "MMg" and arg then
    vGasOff = narg
  elseif data == "MMT2" and arg then
    degOk = narg
  elseif data == "MMT3" and arg then
    deg2Hot = narg
  elseif data == "MMSave" then
	Config_save()		--	in www.lua was local, but we need this global for use from anywhere. (3.5)
  elseif data == "MMTN" and arg then
    local m = "telnet_svr"
    if narg == 0 then
      if telnet_close then
        telnet_close()
        telnet_close = nil
      end
      package.loaded[m] = nil
      _G[m] = nil
    else
      require(m)
    end
  elseif data == "MMFtp" and arg then
    if narg == 0 then
      if FTP then
        FTP.close()
      end
    else
	  local F = require("ftpserver")
	  F.createServer(F, "dev", "dev1234", true)
    end
  else
    send_message("unknown: "..vS)
  end
end

function  MM_doCommand(topic, data)
  MM_doCommandLocal(topic, data)
end



function timer100()
  tick = tick + 1
  if tick >= 100 then
    tick = 0

    -- if bmp then
      -- bmp.read(1)                     -- read the next second.
      -- gTemp = math.floor(bt)  / 10
      -- gHumi = string.format("%4.2f inHg, %4.2f inWC", bp * 75 / 254000, bp * 0.00401)
    -- else
    getTemp()                     -- read the next second.
    -- end

    if vWQ and vWQ > 0 then vWQ = vWQ - 1 end

    vSW = 1 - gpio.read(7)   --pSWCH opposite polarity
    if vSW == 0 then
      if vState == 7 or vState == 1 then	--sFAULT or --sCLR
        vState = 0 -- sOFF
      elseif vState ~= 0 then	-- sOFF
        vState = 6 --sCOOL
      end
    else
      if vState == 0 then	-- sOFF
        vState = 1 --sCLR
      end
    end

    if vStateo ~= vState then
      tkState = 0
    end

    vAdc = adc.read(0)
    gpio.write(5, 1)	--pTHERMLO
    vAdcOpen = adc.read(0)
    gpio.write(5, 0)	--pTHERMLO


    do -- if vAdcOld ~= vAdc then
      local A = vAdc * 1.0 / vAdcOpen
      if A >= .999 then
        -- print(vAdc, vAdcOpen, A)
        A = .999
      end
      local Rt = vR1 * A / (1 - A)
      local vRatio = Rt / vR25
      local degC = calcCent(vRatio)
      -- print(A, Rt, vRatio)
      degCOld = degC1
      degC1 = math.floor(10 * degC + 0.5) / 10
      local degCchg = math.floor(10 * (degC1 - degCOld) + 0.5) / 10
      local Rt0 = math.floor(Rt + 0.5)
	  local vI = vIgnOn > 0 and vIgnOn or vIgn
	  local vG = vGasOff > 0 and -vGasOff or vGas
      MMStatus = "Temp="..degC1.."C ("..degCchg..") at "..vState..":"..tkState
                  .." F="..vFan.." I="..vI.." G="..vG.." S="..vSW.." E="..vStErr.." T="..gTemp.."C H="..gHumi.."% M="..node.heap()
                  .." ("..vAdc.."/"..vAdcOpen..":"..Rt0..") R="..degOk.."~"..deg2Hot
				  .." W="..getWifiStatus()
				  .." B="..vConnect.."."..vWQ.." V=3.5"
      send_message(MMStatus)
      print(MMStatus)
      vAdcOld = vAdc
	  
	  if tkState == 0 then -- save starting temp
	    vStartT = degC1
	  end
    end

    tkState = tkState + 1

    if vState == 0 then	--sOFF
      vFan = 0
      vIgn = 0
      vGas = 0
      vStErr = 0
    elseif vState == 1 then	--sCLR
      vFan = 1
      vIgn = 0
      vGas = 0
      if (tkState >= 30) and (degC1 < degOk) then	--tkCLR
        vState = 2 --sLOW
      end
    elseif vState == 2 then --sLOW
      vFan = 1
      vIgn = 1 -- let the ignitor warm up before turning on the gas
      vGas = 0
      if tkState >= 20 then	-- dev2020-07-01: was 10. let ignitor warm up for 10 seconds before starting gas
        vGas = 1 -- v3.2 was 0
      end
      if tkState >= 35 then	-- dev2020-07-01 was 25 dev2019-09-20 was 15 tkLOW
        vState = 3 --sRISE
      end
    elseif vState == 3 then	--sRISE
      vFan = 1
      vIgn = 1
      vGas = 1
      if degC1 > (vStartT + 5) then	--if degC1 > degOk then
        vState = 4 --sHOT
      elseif tkState >= 180 then	--tkRISE
        vState = 4 --sHOT
      end
    elseif vState == 4 then --sHOT
      vIgn = 0
      if degC1 > (degOk + 1) then
        vState = 5 --sSDY
      elseif tkState >= 300 then	--tkHOT
        vState = 5 --sSDY
      end
    elseif vState == 5 then	--sSDY
      if degC1 < degOk then
        vStErr = 7 --sFAULT
        vState = 6 --sCOOL
	  elseif degC1 > deg2Hot then
        vState = 1 --sCLR
      elseif tkState >= 1814400 then	--tkSDY
        vState = 6 --sCOOL
      end
    elseif vState == 6 then --sCOOL
      vFan = 1
      vIgn = 0
      vGas = 0
      if tkState >= 300 then	--tkCOOL
        if vStErr ~= 0 then	-- v3.2 was vStErr == 7 then --sFAULT
          vState = 7 --sFAULT
        else
          vState = 0 --sOFF
        end
      end
    elseif vState == 7 then --sFAULT
      vFan = 0
      vIgn = 0	-- dev2020-07-13
      -- try to restart -- once per day...
      if tkState >= 86400 then	--tkFAULT
        vState = 1 --sCLR
      end
    end

    if vSW ~= vSWo then
      print("switch = "..vSW.." state = "..vState.." at "..tkState)
      vSWo = vSW
    end
    if vStateo ~= vState then
      print("state = "..vState.." was "..vStateo.." at "..tkState.." ("..vStErr..")")
      vStateo = vState
      tkState = 0
    end

    if vIgnOn > 0 then		-- dev2020-07-02 gpio.write(1, 1) will be overwritten by state 2 code below.
      gpio.write(1, 1)    --pIGN
      vIgnOn = vIgnOn - 1
    else
      gpio.write(1, vIgn)    --pIGN
    end

    if vGasOff > 0 then
      gpio.write(8, 0)    --pGAS
      vGasOff = vGasOff - 1
    else
      gpio.write(8, vGas)	--pGAS
    end
  end

  -- simple 25% duty cycle to reduce fan speed
  if vState == 2 then	--sLOW
    if vFan == 1 then
      if tkFan < 1 then
        gpio.write(6, 1)	--pFAN
      else
        gpio.write(6, 0)	--pFAN
      end

      tkFan = tkFan + 1
      if tkFan >= 4 then
        tkFan = 0
      end
    else
      tkFan = 0
    end

	-- dev2020-07-02 ramp up the ignitor temp
	if tkIgn < 1 then
	  gpio.write(1, 1)
	else
	  gpio.write(1, 0)
	end
	
	tkIgn = tkIgn + 1	-- keep 0 < tkIgn < 20
	if tkIgn >= (20 - tkState) then
	  tkIgn = 0
	end
	
  else
    gpio.write(6, vFan)	--pFAN
  end

  -- blinking light.
  tkLED = tkLED - 1
  if tkLED <= 0 then
    if vStErr ~= 0 then	--sOFF
      tkLED = 12
    else
      tkLED = 50
    end
    if vState == 0 then	--sOFF
      sLed = 0
    elseif vState == 5 then	--sSDY
      sLed = 1
    else
      sLed = 1 - sLed
    end
    --gpio.write(4, sLed)      -- pBLED
    gpio.write(0, 1 - sLed)  -- pLED down is on
  end

end



function module.start()
t6.stop(t6)
t6.alarm(t6, 10, tmr.ALARM_AUTO, timer100)
end

return module
