-- ADS1115 module for ESP8266 with nodeMCU

local moduleName = ...
print(moduleName)
local M = {}
--_G[moduleName] = M

gC = {}		-- 0 = single 0 1 = single 1, 2 - diff 0-1

-- Requirements
--	bit = require("bit")

-- Global variable for i2c
--SDA                                 = 2
--SCL                                 = 3

-- Global variable for ADS1115
local ADDR                                = 0x48

local ADS1115_CONVERTIONDELAY             = 8

local ADS1115_POINTER_MASK              = 0x03
local ADS1115_POINTER_CONVERSION        = 0x00
local ADS1115_POINTER_CONFIG            = 0x01
local ADS1115_POINTER_THRESH_LOW        = 0x02
local ADS1115_POINTER_THRESH_HI         = 0x03


local ADS1115_MUX_MASK            		= 0x7000
-- local ADS1115_MUX_DIFF_0_1        		= 0x0000    -- Differential P = AIN0, N = AIN1 (default)
-- local ADS1115_MUX_DIFF_0_3        		= 0x1000    -- Differential P = AIN0, N = AIN3
-- local ADS1115_MUX_DIFF_1_3        		= 0x2000    -- Differential P = AIN1, N = AIN3
-- local ADS1115_MUX_DIFF_2_3        		= 0x3000    -- Differential P = AIN2, N = AIN3
-- local ADS1115_MUX_SINGLE_0        		= 0x4000    -- Single-ended AIN0
-- local ADS1115_MUX_SINGLE_1        		= 0x5000    -- Single-ended AIN1
-- local ADS1115_MUX_SINGLE_2        		= 0x6000    -- Single-ended AIN2
-- local ADS1115_MUX_SINGLE_3        		= 0x7000    -- Single-ended AIN3

local ADS1115_DR_128SPS           		= 128


local ADS1015_REG_CONFIG_OS_SINGLE        = 0x8000

local ADS1015_REG_CONFIG_CQUE_NONE        = 0x0003
local ADS1015_REG_CONFIG_CLAT_NONLAT      = 0x0000
local ADS1015_REG_CONFIG_CPOL_ACTVLOW     = 0x0000
local ADS1015_REG_CONFIG_CMODE_TRAD       = 0x0000
--local ADS1015_REG_CONFIG_DR_1600SPS       = 0x0080
local ADS1015_REG_CONFIG_MODE_SINGLE      = 0x0100

local ADS1115_DEFAULT_CONFIG_REG 		  = 0x8583    -- Config register value after reset

local ADS1115_PGA_0_256V          		  = 0x0A00    -- +/-0.256V range = Gain 16

-- Functions
local function readRegister(address, reg)
    i2c.start(0)
    i2c.address(0, address, i2c.TRANSMITTER)
    i2c.write(0, reg)
    i2c.stop(0)

    i2c.start(0)
    i2c.address(0, address, i2c.RECEIVER)
    local all = i2c.read(0, 2)
    i2c.stop(0)

    local a = string.byte(all, 1)
    if a == nil then
        a = 0
    end

    local b = string.byte(all, 2)
    if b == nil then
        b = 0
    end

    -- local ret = 0
    -- ret = bit.bor(ret, a)
    -- ret = bit.lshift(ret, 8)
    -- ret = bit.bor(ret, b)
	local ret = a * 256 + b
    return ret
end

local function writeRegister(address, reg, value)
    i2c.start(0)
    i2c.address(0, address, i2c.TRANSMITTER)
    i2c.write(0, reg)
    -- local high = 0
    -- high = bit.rshift(value, 8)
    -- i2c.write(0, high)
    -- local low = 0
    -- low = bit.band(0xFF)
    -- i2c.write(0, low)
	i2c.write(0, value / 256)
	i2c.write(0, value % 256)
    i2c.stop(0)
end


function M.readADCDelay(channel)			--	channel is 0x0000 to 0x7000 see ADS1115_MUX_DIFF_0_1 to 
    local config = ADS1015_REG_CONFIG_CQUE_NONE	+ ADS1015_REG_CONFIG_CLAT_NONLAT + ADS1015_REG_CONFIG_CPOL_ACTVLOW + ADS1015_REG_CONFIG_CMODE_TRAD + ADS1115_DR_128SPS + ADS1015_REG_CONFIG_MODE_SINGLE	
	config = config + ADS1015_REG_CONFIG_OS_SINGLE + channel
	config = config + ADS1115_PGA_0_256V
    writeRegister(ADDR, ADS1115_POINTER_CONFIG, config)
    -------------- bad bad -----------------
	tmr.delay(ADS1115_CONVERTIONDELAY)
    -------------- bad bad -----------------
    local ret = readRegister(ADDR, ADS1115_POINTER_CONVERSION)
    return ret
end

local idxC = 0
local channels = {}

local function getChannelX()
    local ret = readRegister(ADDR, ADS1115_POINTER_CONVERSION)
	if ret > 0x7FFF then
	  ret = ret - 0x10000
	end
    gC[channels[idxC]] = ret
    if idxC < #channels then
	  idxC = idxC + 1
	  M.getChannel()
	end
end

function M.getChannel()
    local config = ADS1015_REG_CONFIG_CQUE_NONE	+ ADS1015_REG_CONFIG_CLAT_NONLAT + ADS1015_REG_CONFIG_CPOL_ACTVLOW + ADS1015_REG_CONFIG_CMODE_TRAD + ADS1115_DR_128SPS + ADS1015_REG_CONFIG_MODE_SINGLE	
    local chan = channels[idxC]
	config = config + ADS1015_REG_CONFIG_OS_SINGLE + chan
    if chan < 0x4000 then
	  config = config + ADS1115_PGA_0_256V
    end
    writeRegister(ADDR, ADS1115_POINTER_CONFIG, config)
    tmr.create():alarm(10, tmr.ALARM_SINGLE, getChannelX)
end

function M.init(sda, scl)
  --print('I2C_SETUP...')
  i2c.setup(0, sda, scl, i2c.SLOW)	-- i2c.setup(0, SDA, SCL, i2c.SLOW)
  --print('SET UP')

	-- static const uint8_t general_i2c_addr = 0x00;
	-- static const uint8_t ads1115_i2c_reset = 0x06;
  -- writeRegister(ADDR, 0, 0x06)
    i2c.start(0)
    i2c.address(0, 0, i2c.TRANSMITTER)
    i2c.write(0, 0x06)
    i2c.stop(0)
  
  local config_read = readRegister(ADDR, ADS1115_POINTER_CONFIG)
  print(config_read, ADS1115_DEFAULT_CONFIG_REG)
  if config_read == 0xffff then
    return -1
  end
  return 0
end


function M.testDelay()
--    local a = M.readADCDelay(0x4000)	-- 0
--    local b = M.readADCDelay(0x5000)	-- 1
--    local c = M.readADCDelay(0x0000)	-- 0 - 1
--    print("0,1,0-1: "..a.." - "..b.." = "..c)
    for chan, raw in pairs(gC) do 
	  local mV = 0
	  if chan < 0x4000 then
	    mV = raw * 0.0078125
	  else
	    mV = raw * 0.1875
	  end
	  print(chan, raw, mV)
	end
	idxC = 1
	channels ={ 0x4000, 0x5000, 0x0000 }
    --channels ={ 0x0000 }
    M.getChannel()
end

--tmr.alarm(0, 1000, 1, function() main() end )
local ret = M.init(3, 4)	-- SDA = 3, SCL = 4
if ret ~= 0 then
    print("device not found")
else
  tc = tmr.create()
  tmr.alarm(tc, 1000, tmr.ALARM_AUTO, M.testDelay)
end

return M
