-- ***************************************************************************
-- BME280 module for ESP8266 with nodeMCU
--
-- requires delay after init for valid values. 
--
-- MIT license, http://opensource.org/licenses/MIT
-- ***************************************************************************

-- local moduleName = ...
local M = {}
-- _G[moduleName] = M

-- bmp28eDbg = 0

local function getAddr(dev)
  if dev == 1 then
    return 0x77
  end
  return 0x76
end

local BME280_REGISTER_CONTROL_HUM = 0xF2
local BME280_REGISTER_CONTROL = 0xF4
local BME280_REGISTER_CONFIG  = 0xF5
local BME280_REGISTER_CHIPID  = 0xD0

local BME280_REGISTER_T       = 0x88 --	0x88-0x8D ( 6)
local BME280_REGISTER_P       = 0x8E --	0x8E-0x9F (18)
local BME280_REGISTER_H1      = 0xA1 --	0xA1      ( 1)
local BME280_REGISTER_H2      = 0xE1 --	0xE1-0xE7 ( 7)
local BME280_REGISTER_PRESS   = 0xF7 -- 0xF7-0xF9

-- i2c interface ID
local id = 0

-- 16-bit  two's complement
-- value: 16-bit integer
local function twoCompl(value)
 if value > 32767 then 
   value = -(65535 - value + 1)
 end
 return value
end

local function twoCompl12(value)
 if value > 0x0fff then 
   value = -(0x7ff - value + 1)
 end
 return value
end

-- read data register
-- reg_addr: address of the register in BMP180
-- lenght: bytes to read
local function read_reg(ADDR, reg_addr, length)
  i2c.start(id)
  i2c.address(id, ADDR, i2c.TRANSMITTER)
  i2c.write(id, reg_addr)
  i2c.stop(id)
  i2c.start(id)
  i2c.address(id, ADDR,i2c.RECEIVER)
  c = i2c.read(id, length)
  i2c.stop(id)
  return c
end

-- write data register
-- reg_addr: address of the register in BMP180
-- reg_val: value to write to the register
local function write_reg(ADDR, reg_addr, reg_val)
  i2c.start(id)
  i2c.address(id, ADDR, i2c.TRANSMITTER)
  i2c.write(id, reg_addr)
  i2c.write(id, reg_val)
  i2c.stop(id)
end


function createDevice(sda, scl, dev)
	local isInit = false
	local T1, T2, T3, P1, P2, P3, P4, P5, P6, P7, P8, P9, H1, H2, H3, H4, H5, H6
	local ADDR = 0
	local bme280_isbme = false
	local chipid = 0

	-- initialize module
	-- sda: SDA pin
	-- scl SCL pin
	function init()		--	sda, scl, devX)	use upvalue
	  i2c.setup(id, sda, scl, i2c.SLOW)
	  ADDR = getAddr(dev)
	  print("BME280 init", sda, scl, dev, ADDR)
		-- uint8_t const bit3 = 0b111;
		-- uint8_t const bit2 = 0b11;

		-- bme280_mode = (!lua_isnumber(L, 4)?BME280_NORMAL_MODE:(luaL_checkinteger(L, 4)&bit2)) // 4-th parameter: power mode
			-- | ((!lua_isnumber(L, 2)?BME280_OVERSAMP_16X:(luaL_checkinteger(L, 2)&bit3)) << 2) // 2-nd parameter: pressure oversampling
			-- | ((!lua_isnumber(L, 1)?BME280_OVERSAMP_16X:(luaL_checkinteger(L, 1)&bit3)) << 5); // 1-st parameter: temperature oversampling
			
		-- bme280_ossh = (!lua_isnumber(L, 3))?BME280_OVERSAMP_16X:(luaL_checkinteger(L, 3)&bit3); // 3-rd parameter: humidity oversampling
		
		-- config = ((!lua_isnumber(L, 5)?BME280_STANDBY_TIME_20_MS:(luaL_checkinteger(L, 5)&bit3))<< 5) // 5-th parameter: inactive duration in normal mode
			-- | ((!lua_isnumber(L, 6)?BME280_FILTER_COEFF_16:(luaL_checkinteger(L, 6)&bit3)) << 2); // 6-th parameter: IIR filter
		-- full_init = !lua_isnumber(L, 7)?1:lua_tointeger(L, 7); // 7-th parameter: init the chip too
		-- NODE_DBG("mode: %x\nhumidity oss: %x\nconfig: %x\n", bme280_mode, bme280_ossh, config);

	  --if (full_init) 
	  
	  local cid = read_reg(ADDR, BME280_REGISTER_CHIPID, 1)
	  chipid = string.byte(cid,1)
	  -- print(cid, chipid)
	  bme280_isbme = chipid == 0x60
	  
	  if bme280_isbme == true then
	    print("enabling humidity")
	    local bme280_ossh = 7		-- x8 oversampling.
	    write_reg(ADDR, BME280_REGISTER_CONTROL_HUM, bme280_ossh)	-- changes effective only write to control_meas
	  end

	  local bme280_mode = 3		-- normal mode
	  bme280_mode = bme280_mode + 0xFC			--	oversampling temperature and pressure to the max.
	  write_reg(ADDR, BME280_REGISTER_CONTROL, bme280_mode)

	  
	  -- lsb, msb  
	  local calBytes = read_reg(ADDR, BME280_REGISTER_T, 24)
		local outstr = ""
		for i = 1, #calBytes do
		  outstr = outstr .. string.format(" %02x", string.byte(calBytes, i))
		end
		print(outstr)

	  T1 = string.byte(calBytes, 2) * 256 + string.byte(calBytes, 1)			--	unsigned short
	  T2 = twoCompl(string.byte(calBytes, 4) * 256 + string.byte(calBytes, 3))
	  T3 = twoCompl(string.byte(calBytes, 6) * 256 + string.byte(calBytes, 5))
	  P1 = string.byte(calBytes, 8) * 256 + string.byte(calBytes, 7)			--	unsigned short
	  P2 = twoCompl(string.byte(calBytes, 10) * 256 + string.byte(calBytes, 9))
	  P3 = twoCompl(string.byte(calBytes, 12) * 256 + string.byte(calBytes, 11))
	  P4 = twoCompl(string.byte(calBytes, 14) * 256 + string.byte(calBytes, 13))
	  P5 = twoCompl(string.byte(calBytes, 16) * 256 + string.byte(calBytes, 15))
	  P6 = twoCompl(string.byte(calBytes, 18) * 256 + string.byte(calBytes, 17))
	  P7 = twoCompl(string.byte(calBytes, 20) * 256 + string.byte(calBytes, 19))
	  P8 = twoCompl(string.byte(calBytes, 22) * 256 + string.byte(calBytes, 21))
	  P9 = twoCompl(string.byte(calBytes, 24) * 256 + string.byte(calBytes, 23))
	  
	  print(T1, T2, T3, P1, P2, P3, P4, P5, P6, P7, P8, P9)
	  
	  if bme280_isbme == true then
		H1 = read_reg(ADDR, BME280_REGISTER_H1, 1)
		H1 = string.byte(H1, 1)
		calBytes = read_reg(ADDR, BME280_REGISTER_H2, 7)

		outstr = string.format(" %02x", H1)
		for i = 1, #calBytes do
		  outstr = outstr .. string.format(" %02x", string.byte(calBytes, i))
		end
		print(outstr)

		H2 = twoCompl(string.byte(calBytes, 2) * 256 + string.byte(calBytes, 1))
		H3 = string.byte(calBytes, 3)
		--H4 = (int16_t)reg[0] << 4 | (reg[1] & 0x0F); reg+=1;	// H4[11:4 3:0] = 0xE4[7:0] 0xE5[3:0] 12-bit signed
		H4 = string.byte(calBytes, 4) * 16 + string.byte(calBytes, 5) % 16
		--H5 = (int16_t)reg[1] << 4 | (reg[0]   >> 4); reg+=2;	// H5[11:4 3:0] = 0xE6[7:0] 0xE5[7:4] 12-bit signed
		H4 = twoCompl12(H4)
		H5 = string.byte(calBytes, 6) * 16 + math.floor(string.byte(calBytes, 5) / 16)
		H5 = twoCompl12(H5)
		H6 = string.byte(calBytes, 7)
		print(H1, H2, H3, H4, H5, H6)
	  end
	  isInit = true
	end

	function read_temp()
		local t_fine = 0
		local bp = 0
		local bt = 0
		local bh = 0

		function bme280_compensate_T(adc_T)
		--  print("hey")
		--  print(T1, T2, T3, P1, P2, P3, P4, P5, P6, P7, P8, P9)
		  local var1 = (adc_T / 16384.0 - T1 / 1024.0) * T2
		--  print(adc_T, t_fine, var1)
		  local var2 = (adc_T / 131072.0 - T1 / 8192.0) * (adc_T / 131072.0 - T1 / 8192.0) * T3
		--  print(adc_T, t_fine, var1, var2)
		  t_fine = var1 + var2
		--  print(adc_T, t_fine, var1, var2)
		  local T = t_fine / 5120.0
		  return T;
		end

		function bme280_compensate_P(adc_P)
		  local var1 = t_fine / 2.0 - 64000.0
		  local var2 = var1 * var1 * P6 / 32768.0
		  var2 = var2 + var1 * P5 * 2.0
		  var2 = var2 / 4.0 + P4 * 65536.0
		  var1 = (P3 * var1 * var1 / 524288.0 + P2 * var1) / 524288.0
		  var1 = (1.0 + var1 / 32768.0) * P1
		  if var1 == 0 then
			return 0	-- avoid exception caused by division by zero
		  end
		  local p = 1048576.0 - adc_P
		  p = (p - var2 / 4096.0) * 6250.0 / var1
		  var1 = P9 * p * p / 2147483648.0
		  var2 = p * P8 / 32768.0
		  p = p + (var1 + var2 + P7) / 16.0
		  return p;
		end

		function bme280_compensate_H(adc_H) 
			local val = t_fine * 1.0 - 76800.0
		--	print(adc_H, val, t_fine, H1, H2, H3, H4, H5, H6)
			val = (adc_H - (H4 * 64.0 + H5 / 16384.0 * val)) 
			      * (H2 / 65536.0 * (1.0 + H6 / 67108864.0 * val * (1.0 + H3 / 67108864.0 * val)))
		--	print(val)
            val = val * (1.0 - H1 * val / 524288.0)
		--	print(val)
			if val > 100.0 then
			  val = 100.0
			end
			if val < 0.0 then
			  val = 0.0
			end
			return val
		end 


  	   local buf = read_reg(ADDR, BME280_REGISTER_PRESS, 8)	--	registers are P[3], T[3], H[2]
	  
		-- Must do Temp first since bme280_t_fine is used by the other compensation functions
		-- whole thing is x16	
		local adc_T = (string.byte(buf, 4) * 65536 + string.byte(buf, 5) * 256 + string.byte(buf, 6)) / 16
	--  print(adc_T)
	--	print(t_fine)
		if adc_T == 0x80000 or adc_T == 0xfffff then
			return
		end		
		bt = bme280_compensate_T(adc_T)
	--  print(adc_T, t_fine)

		local adc_P = (string.byte(buf, 1) * 65536 + string.byte(buf, 2) * 16 + string.byte(buf, 3)) / 16
		if adc_P == 0x80000 or adc_P == 0xfffff then 
			bp = 0
			-- calc_qnh = 0;
		else
			bp = bme280_compensate_P(adc_P)	-- 
		end

		local adc_H = string.byte(buf, 7) * 256 + string.byte(buf, 8)
		-- print(string.format("%x", adc_H), bme280_isbme, string.format("%02x", chipid))
		if (bme280_isbme == false or adc_H == 0x8000 or adc_H == 0xffff) then
			bh = 0
		else
			bh = bme280_compensate_H(adc_H)
		end
--[[
		if calc_qnh == 1 then 		-- have altitude
			int32_t h = luaL_checkinteger(L, 1);
			double qnh = bme280_qfe2qnh(qfe, h);
			lua_pushinteger(L, (int32_t)(qnh + 0.5));
			return 4;
		end
]]
--[[		
	  if (bmp28eDbg > 2) then
		print("TempX", dev, ADDR)

		local outstr = ""
		for i = 1, #buf do
		  outstr = outstr .. string.format(" %02x", string.byte(buf, i))
		end
		print(outstr)
		print(T1, T2, T3, P1, P2, P3, P4, P5, P6, P7, P8, P9)
		 math.floor(bt * 100) / 100)
		print(adc_T, t_fine, string.format("%x", adc_T), string.format("%4.2f", bt))
		pinHg = 0.02952998751 × phPa
		print(adc_P, string.format("%x", adc_P), bp, bp * 0.00029529988)
	--	print(adc_H, string.format("%x", adc_H), bh)
	  end
]]
	  return bt, bp, bh
	end

	if isInit == false then
		if sda == nil or scl == nil or dev == nil then
		  return nil
		end
		init()	-- use upvalues sda, scl, dev)
	end
	return read_temp
end


--[[
bmp28eDbg = 3
if bmp28eDbg > 2 then
  local function nh(s)
	print(s,node.heap())
  end

  local pDev, pAlt

  local function testDelay()
	local lbt, lbp, lbh = pDev()
	print(lbt, lbp, lbp * 0.00029529988, lbh)

	lbt, lbp, lbh = pAlt()
	print(lbt, lbp, lbp * 0.00029529988, lbh)
  end


  nh("0")
  pDev = createDevice(3, 4, 2)	-- SDA = 3, SCL = 4
  print(pDev)
  nh("1")
  pAlt = createDevice(3, 4, 1)	-- SDA = 3, SCL = 4
  print(pAlt)
  nh("2")
  tc = tmr.create()
  if bmp28eDbg > 3 then
	tc.alarm(tc, 3000, tmr.ALARM_AUTO, testDelay)
  else
	tc.alarm(tc, 1000, tmr.ALARM_SINGLE, testDelay)
  end
  nh("3")
end
]]

return M
